import os.path
import shutil
import time

import matplotlib.pyplot as plt
import numpy as np
import pickle
import arithmetics
from graph import act_t, Node, NodePtr

np.set_printoptions(precision=4, suppress=True, linewidth=np.inf)

COLORS = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']


class Goofspiel(object):
    def __init__(self):
        # non-iterables
        self.CARD_NUM: int = 4
        self.upcard: list[int] = [1, 1, 1, 1, 1]
        self.weight: dict[str, list[float]] = {"A": [5.00, 1.33, 2.71, 1.80, 2.24], "B": [4.12, 6.28, 3.33, 1.92, 3.60]}
        self.node_layer: list[list[NodePtr]] = []  # nodes in breadth-first layout

        # states
        self.reg: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # reg[play_name][info][act]
        self.sig: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # sig[play_name][info][act]
        self.opt: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # opt[play_name][info][act]
        self.tab: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # tab[play_name][state][act, act]
        self.cfv: dict[NodePtr, np.ndarray] = {}  # cfv[node][play_id]
        self.exp: dict[NodePtr, np.ndarray] = {}  # exp[state][play_id]

        # records
        self.REGs: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # REGs[play][info][act, t]
        self.SIGs: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # SIGs[play][info][act, t]
        self.OPTs: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # OPTs[play][info][act, t]
        self.CFVs: dict[NodePtr, np.ndarray] = {}  # CFVs[state][play, t]
        self.EXPs: dict[NodePtr, np.ndarray] = {}  # EXPs[state][play, t]

        self.node_layer_init()
        self.val_init()
        self.leaf_assign()

        with open("node.db", "wb") as db:
            pickle.dump(Node.MEM_N, db)
            pickle.dump(Node.MEM, db)
            pickle.dump(self.CARD_NUM, db)
            pickle.dump(self.upcard, db)
            pickle.dump(self.weight, db)
            pickle.dump(self.node_layer, db)

        with open("state.db", "wb") as db:
            pickle.dump(self.reg, db)

    def node_layer_init(self):
        root = Node.new()
        root.o.branch = np.array([0, 0])
        root.o.actA = np.arange(self.CARD_NUM) + 1
        root.o.actB = np.arange(self.CARD_NUM) + 1

        parent = [root]
        child = []
        self.node_layer = []
        while len(parent):
            self.node_layer.append(parent)
            for n in parent:
                n.o.child = np.resize(n.o.child, [len(n.o.actA), len(n.o.actB)])
                for i, a in enumerate(n.o.actA):
                    for j, b in enumerate(n.o.actB):
                        n.o.child[i, j] = Node.new()
                        ch = n.o.child[i, j]
                        ch.o.parent = n
                        ch.o.branch = np.array([a, b])
                        ch.o.actA = n.o.actA[n.o.actA != a]
                        ch.o.actB = n.o.actB[n.o.actB != b]
                        child.append(ch)
            parent = child
            child = []

    def val_init(self):
        for layer in self.node_layer[0:-2]:
            for n in layer:
                self.reg["A"][n] = arithmetics.rand_f(1, len(n.o.actA) - 1)[0] * 5
                self.reg["B"][n] = arithmetics.rand_f(1, len(n.o.actB) - 1)[0] * 5
                self.sig["A"][n] = np.empty([len(n.o.actA)], dtype=float)
                self.sig["B"][n] = np.empty([len(n.o.actB)], dtype=float)
                self.tab["A"][n] = np.empty([len(n.o.actA), len(n.o.actB)], dtype=float)
                self.tab["B"][n] = np.empty([len(n.o.actA), len(n.o.actB)], dtype=float)
                self.opt["A"][n] = np.empty([len(n.o.actA)], dtype=float)
                self.opt["B"][n] = np.empty([len(n.o.actB)], dtype=float)
                self.cfv[n] = np.empty([2], dtype=float)
                self.exp[n] = np.empty([2], dtype=float)
        for layer in self.node_layer[-2:]:
            for n in layer:
                self.cfv[n] = np.empty([2], dtype=float)

    def leaf_assign(self):
        for n in self.node_layer[-1]:
            h = n.o.h()
            u, v = 0, 0
            for k, acts in enumerate(h):
                a, b = acts
                if a > b:
                    u += self.upcard[k] * self.weight["A"][k]
                    v -= self.upcard[k] * self.weight["B"][k]
                elif a < b:
                    u -= self.upcard[k] * self.weight["A"][k]
                    v += self.upcard[k] * self.weight["B"][k]
            self.cfv[n][:] = np.array([u, v], dtype=float)
        for n in self.node_layer[-2]:
            self.cfv[n][:] = self.cfv[n.o.get_child(0, 0)].copy()

    def log_init(self, T=501):
        for layer in self.node_layer[0:-2]:
            for n in layer:
                self.REGs["A"][n] = np.empty([len(n.o.actA), T], dtype=float)
                self.REGs["B"][n] = np.empty([len(n.o.actB), T], dtype=float)
                self.SIGs["A"][n] = np.empty([len(n.o.actA), T], dtype=float)
                self.SIGs["B"][n] = np.empty([len(n.o.actB), T], dtype=float)
                self.OPTs["A"][n] = np.empty([len(n.o.actA), T], dtype=float)
                self.OPTs["B"][n] = np.empty([len(n.o.actB), T], dtype=float)
                self.CFVs[n] = np.empty([2, T], dtype=float)
                self.EXPs[n] = np.empty([2, T], dtype=float)

    def reg_upd(self):
        for layer in self.node_layer[0:-2]:
            for n in layer:
                x = self.sig["A"][n][:]
                y = self.sig["B"][n][:]
                u = self.opt["A"][n][:]
                v = self.opt["B"][n][:]

                r = {"A": u - u.T @ x, "B": v - v.T @ y}
                r["A"][r["A"] < 0] *= 0.20
                r["B"][r["B"] < 0] *= 0.20
                self.reg["A"][n][:] = self.reg["A"][n][:] + r["A"]
                self.reg["B"][n][:] = self.reg["B"][n][:] + r["B"]
                self.reg["A"][n][self.reg["A"][n][:] < 0] = 0
                self.reg["B"][n][self.reg["B"][n][:] < 0] = 0

    def sync(self):
        for layer in self.node_layer[-3::-1]:
            for n in layer:
                self.sig["A"][n][:] = self.reg["A"][n][:] / self.reg["A"][n][:].sum()
                self.sig["B"][n][:] = self.reg["B"][n][:] / self.reg["B"][n][:].sum()
                for i, a in enumerate(n.o.actA):
                    for j, b in enumerate(n.o.actB):
                        self.tab["A"][n][i, j], self.tab["B"][n][i, j] = self.cfv[n.o.get_child(i, j)][:]
                self.opt["A"][n][:] = self.tab["A"][n] @ self.sig["B"][n]
                self.opt["B"][n][:] = self.tab["B"][n].T @ self.sig["A"][n]
                self.cfv[n][0] = self.opt["A"][n].T @ self.sig["A"][n]
                self.cfv[n][1] = self.opt["B"][n].T @ self.sig["B"][n]
                self.exp[n][0] = self.opt["A"][n][:].max() - self.cfv[n][0]
                self.exp[n][1] = self.opt["B"][n][:].max() - self.cfv[n][1]

    def log_rec(self, t: int):
        for layer in self.node_layer[-3::-1]:
            for n in layer:
                self.REGs["A"][n][:, t] = self.reg["A"][n][:].copy()
                self.REGs["B"][n][:, t] = self.reg["B"][n][:].copy()
                self.SIGs["A"][n][:, t] = self.sig["A"][n][:].copy()
                self.SIGs["B"][n][:, t] = self.sig["B"][n][:].copy()
                self.OPTs["A"][n][:, t] = self.opt["A"][n][:].copy()
                self.OPTs["B"][n][:, t] = self.opt["B"][n][:].copy()
                self.CFVs[n][:, t] = self.cfv[n].copy()
                self.EXPs[n][:, t] = self.exp[n].copy()


def run(game: Goofspiel, T=501):
    # values that constant to iterations
    with open("node.db", "rb") as db:
        Node.MEM_N = pickle.load(db)
        Node.MEM = pickle.load(db)
        game.CARD_NUM = pickle.load(db)
        game.upcard = pickle.load(db)
        game.weight = pickle.load(db)
        game.node_layer = pickle.load(db)
        game.leaf_assign()
    # state
    with open("state.db", "rb") as db:
        game.reg = pickle.load(db)

    game.log_init(T)
    t_start = time.time()
    for t in range(0, T):
        if t % 50 == 0:
            print(t)
        game.sync()
        game.log_rec(t)
        game.reg_upd()
    print("iter:", time.time() - t_start)

    with open("state.db", "wb") as db:
        pickle.dump(game.reg, db)

    if "logs" not in os.listdir():
        os.mkdir("logs")
    with open(os.path.join("logs", "REGs.db"), "wb") as db:
        pickle.dump(game.REGs, db)
    with open(os.path.join("logs", "SIGs.db"), "wb") as db:
        pickle.dump(game.SIGs, db)
    with open(os.path.join("logs", "OPTs.db"), "wb") as db:
        pickle.dump(game.OPTs, db)
    with open(os.path.join("logs", "CFVs.db"), "wb") as db:
        pickle.dump(game.CFVs, db)
    with open(os.path.join("logs", "EXPs.db"), "wb") as db:
        pickle.dump(game.EXPs, db)


def split_to_stackers(A: np.ndarray):
    M, N = A.shape
    y = []
    for i in range(M):
        y.append(A[i, :])
    return y


def output(game: Goofspiel, T=501):
    try:
        shutil.rmtree("curves")
    except:
        pass
    finally:
        os.mkdir("curves")
        for d in ["cfv_A", "cfv_B", "ep", "reg_A", "reg_B", "sig_A", "sig_B"]:
            os.mkdir(os.path.join("curves", d))
    # reg
    t_start = time.time()
    for l in game.node_layer[0:-3]:
        for n in l:
            h = str(n.o.h())
            fig = plt.figure(h, [8, 4], 96)

            ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
            colors = [COLORS[a % 10] for a in n.o.actA]
            ax.stackplot(np.arange(T), *split_to_stackers(game.REGs["A"][n]), labels=n.o.actA, colors=colors, alpha=0.5)
            ax.set_title(r'$R^A$' + h)
            ax.grid()
            ax.legend()
            fig.savefig(os.path.join("curves", "reg_A", h + ".png"))

            ax.clear()
            colors = [COLORS[b % 10] for b in n.o.actB]
            ax.stackplot(np.arange(T), *split_to_stackers(game.REGs["B"][n]), labels=n.o.actB, colors=colors, alpha=0.5)
            ax.set_title(r'$R^B$' + h)
            ax.grid()
            ax.legend()
            fig.savefig(os.path.join("curves", "reg_B", h + ".png"))
            plt.close(fig)
    print("reg:", time.time() - t_start)

    # sig
    t_start = time.time()
    for l in game.node_layer[0:-3]:
        for n in l:
            h = str(n.o.h())
            fig = plt.figure(h, [8, 4], 96)
            ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])

            colors = [COLORS[a % 10] for a in n.o.actA]
            ax.stackplot(np.arange(T), *split_to_stackers(game.SIGs["A"][n]), labels=n.o.actA, colors=colors, alpha=0.5)
            ax.set_title(r'$\sigma^A$' + h)
            ax.grid()
            ax.legend()
            fig.savefig(os.path.join("curves", "sig_A", h + ".png"))

            ax.clear()
            colors = [COLORS[b % 10] for b in n.o.actB]
            ax.stackplot(np.arange(T), *split_to_stackers(game.SIGs["B"][n]), labels=n.o.actB, colors=colors, alpha=0.5)
            ax.set_title(r'$\sigma^B$' + h)
            ax.grid()
            ax.legend()
            fig.savefig(os.path.join("curves", "sig_B", h + ".png"))
            plt.close(fig)
    print("sig:", time.time() - t_start)

    # cfv
    t_start = time.time()
    for l in game.node_layer[0:-3]:
        for n in l:
            h = str(n.o.h())
            fig = plt.figure(h, [8, 4], 96)
            ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
            ax.plot(np.arange(T), game.CFVs[n][0], label="mixed")
            for i, a in enumerate(n.o.actA):
                ax.plot(np.arange(T), game.OPTs["A"][n][i, :], label=a, alpha=0.3, color="C" + str(a))
            ax.set_title(r'$cfv^A$' + h)
            ax.grid()
            ax.legend()
            fig.savefig(os.path.join("curves", "cfv_A", h + ".png"))

            ax.clear()
            ax.plot(np.arange(T), game.CFVs[n][1], label="mixed")
            for j, b in enumerate(n.o.actB):
                ax.plot(np.arange(T), game.OPTs["B"][n][j, :], label=b, alpha=0.3, color="C" + str(b))
            ax.set_title(r'$cfv^B$' + h)
            ax.grid()
            ax.legend()
            fig.savefig(os.path.join("curves", "cfv_B", h + ".png"))
            plt.close(fig)
    print("cfv:", time.time() - t_start)

    # ep-value
    t_start = time.time()
    for l in game.node_layer[0:-3]:
        for n in l:
            h = str(n.o.h())
            fig: plt.Figure = plt.figure(h, [8, 4], 96)
            fig.add_axes([0.1, 0.1, 0.8, 0.8])
            fig.axes[0].loglog(np.arange(T), game.EXPs[n].sum(axis=0), label=r'$ep$')
            fig.axes[0].loglog(np.arange(T), game.EXPs[n][0, :], label=r'$ep_A$', alpha=0.3)
            fig.axes[0].loglog(np.arange(T), game.EXPs[n][1, :], label=r'$ep_B$', alpha=0.3)
            fig.axes[0].set_title(r'$ep$' + h)

            ax: plt.Axes
            for ax in fig.axes:
                ax.set_ylim(0.01, 10)
                ax.grid()
                ax.legend()
            fig.savefig(os.path.join("curves", "ep", h + ".png"))
            plt.close(fig)
    print("exp:", time.time() - t_start)


if __name__ == "__main__":
    np.random.seed(133484)
    game = Goofspiel()
    run(game, 501)
    output(game, 501)
